from typing import List, Tuple
from Bio.PDB.Atom import Atom
import numpy as np
from scipy.spatial.transform import Rotation as R
from scipy.spatial.distance import pdist


class CoordManipulator:
    def __init__(self, entity) -> None:
        self.entity = entity
        self._atoms, self.coords = self._extract_atoms_and_coords(entity)

    def _extract_atoms_and_coords(self, entity) -> Tuple[List[Atom], np.array]:
        coords = []
        atoms = []
        for atom in entity.get_atoms():
            coords.append(atom.coord)
            atoms.append(atom)
        return atoms, np.asarray(coords)

    def get_dist_matrix(self) -> np.array:
        if not hasattr(self, "_dist_matrix"):
            self._dist_matrix = pdist(self.coords)
        return self._dist_matrix

    def get_farthest_atom_pair(self) -> Tuple[Atom, Atom]:
        a1, a2 = self.get_farthest_atom_indices()
        # return the atom objects themselves, not the indices
        return self._atoms[a1], self._atoms[a2]

    def get_farthest_atom_indices(self) -> Tuple[int, int]:
        if not hasattr(self, "_a1"):
            self._a1, self._a2 = self._find_farthest_atom_indices()
        # return the atom indices in the dist_mat
        return self._a1, self._a2

    def _find_farthest_atom_indices(self) -> Tuple[int, int]:
        dists = self.get_dist_matrix()
        idx_pair = np.unravel_index(
            np.argmax(dists, axis=None),
            dists.shape
        )
        return idx_pair

    def get_transformation_matrices(self) -> Tuple[np.array, np.array]:
        if not hasattr(self, "m_translation"):
            self._find_transformation_operators()
        return self.m_translation, self.m_rotation

    def _find_transformation_operators(self) -> None:
        dists = self.get_dist_matrix()
        a1, a2 = self.get_farthest_atom_indices()
        # Translation operator is just by moving a1 to origin (0, 0, 0)
        m_translation = -self.coords[a1]
        # Find the coord of a2 after the translation
        vec_translated = (self.coords[a2] + m_translation).reshape(1,-1)
        # Define the final coord of a2 after the rotation (dist(a1,a2), 0, 0),
        # that is, align a1, a2 on the x-axis
        vec_rotated = np.array([dists[a1, a2],0,0]).reshape(1,-1)
        # Estimate the rotation by Kabsch algorithm implemented in scipy
        rot_obj, rssd = R.align_vectors(vec_translated, vec_rotated)
        # Get the rotation operator as a numpy array
        m_rotation = rot_obj.as_matrix()
        # Test if the translation and rotation operators can
        # transform a2 to the final coord within error
        assert np.allclose(
            (self.coords[a2] + m_translation) @ m_rotation,
            vec_rotated
        )
        self.m_translation, self.m_rotation = m_translation, m_rotation
    
    def transform_coords(self, coords) -> np.array:
        translation, rotation = self.get_transformation_matrices()
        new_coords = (coords + translation) @ rotation
        return new_coords

    def apply_transformation(self) -> None:
        for i, atom in enumerate(self._atoms):
            atom.coord = self.coords[i]

    def orient_coords(self) -> None:
        self._find_transformation_operators()
        self.coords = self.transform_coords(self.coords)
        self.apply_transformation()





from crimm.StructEntities.TopoElements import Bond, Angle, Dihedral, Improper
from crimm.Modeller import TopologyLoader, ParameterLoader, SeqChainGenerator
from crimm.Modeller.TopoFixer import ResidueFixer

class Topology:
    def __init__(self) -> None:
        self.definitions = {}
        self.params = {}
        self._1to3 = None
        self._3to1 = None
        self._report_res = None
        self.resnames = None
        self.built_residues = None
    
    def load_definition(self, entity_type: str):
        definition = TopologyLoader(entity_type)
        param = ParameterLoader(entity_type)
        param.fill_ic(definition)
        self.definitions[entity_type] = definition



import os
import warnings
from crimm.IO.PRMParser import categorize_lines, parse_line_dict, skip_line
from crimm.Modeller.TopoLoader import TopologyElementContainer
import crimm.StructEntities as Entities

toppar_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../Data/toppar')
)
prm_path_dict = {
    "protein": os.path.join(toppar_dir, 'prot.prm'),
    "nucleic": os.path.join(toppar_dir, 'na.prm'),
    "lipid": os.path.join(toppar_dir, 'lipid.prm'),
    "carb": os.path.join(toppar_dir, 'carb.prm'),
    "ethers": os.path.join(toppar_dir, 'ethers.prm'),
    "cgenff": os.path.join(toppar_dir, 'cgenff.prm')
}


class ParameterLoader:
    ic_position_dict = {
        'R(I-J)': (0, 1),
        'T(I-J-K)': (0, 1, 2),
        'T(J-K-L)': (1, 2, 3),
        'R(K-L)': (2, 3),
        'T(I-K-J)': (0, 2, 1),
        'R(I-K)': (0, 2),
    }
    """A dictionary that stores parameters for CHARMM force field."""
    def __init__(self, file_path=None):
        self.param_dict = {}
        self._raw_data_strings = []
        if file_path is not None:
            self.load_type(file_path)

    def load_type(self, entity_type:str):
        """Load parameters from a CHARMM prm file."""
        entity_type = entity_type.lower()
        if entity_type not in prm_path_dict:
            raise ValueError(f'No parameter file for {entity_type}')
        filename = prm_path_dict[entity_type]
        with open(filename, 'r', encoding='utf-8') as f:

            self._raw_data_strings = [
                l.rstrip() for l in f.readlines() #if not skip_line(l)
            ]
            param_line_dict = categorize_lines(self._raw_data_strings)
        self.param_dict.update(parse_line_dict(param_line_dict))

    def __repr__(self):
        n_bonds = len(self.param_dict['bonds'])
        n_angles = len(self.param_dict['angles'])
        n_urey_bradley = len(self.param_dict['urey_bradley'])
        n_dihedrals = len(self.param_dict['dihedrals'])
        n_impropers = len(self.param_dict['improper'])
        n_cmaps = len(self.param_dict['cmap'])
        n_nonbonds = len(self.param_dict['nonbonded'])
        n_nonbond14s = len(self.param_dict['nonbonded14'])
        n_nbfixes = len(self.param_dict['nbfix'])
        return (
            f'<ParameterDict Bond: {n_bonds}, Angle: {n_angles}, '
            f'Urey Bradley: {n_urey_bradley}, Dihedral: {n_dihedrals}, '
            f'Improper: {n_impropers}, CMAP: {n_cmaps}, '
            f'Nonbond: {n_nonbonds}, Nonbond14: {n_nonbond14s}, '
            f'NBfix: {n_nbfixes}>'
        )

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key):
        return self.param_dict[key]

    def _get_param(self, param_dict: dict, key):
        return (
            param_dict.get(key) or param_dict.get(tuple(reversed(key)))
        )

    def _get_from_choices(self, param_dict: dict, matching_orders: tuple):
        for choice in matching_orders:
            if value := self._get_param(param_dict, choice):
                return value

    def get_bond(self, key):
        """Get bond parameters for a given bond instance."""
        bond_dict = self.param_dict['bonds']
        return self._get_param(bond_dict, key)

    def get_angle(self, key):
        """Get angle parameters for a given angle instance."""
        angle_dict = self.param_dict['angles']
        return self._get_param(angle_dict, key)

    def get_dihedral(self, key):
        """Get dihedral parameters for a given dihedral instance."""
        A, B, C, D = key
        matching_orders = (
            (A, B, C, D),
            ('X', B, C, 'X'),
        )
        dihedral_dict = self.param_dict['dihedrals']
        return self._get_from_choices(dihedral_dict, matching_orders)

    def get_improper(self, key):
        """Get improper parameters for a given improper instance."""
        A, B, C, D = key
        matching_orders = (
            ( A,   B,   C,  D ),
            ( A,  'X', 'X', D ),
            ('X',  B,   C,  D ),
            ('X',  B,   C, 'X'),
            ('X', 'X',  C,  D )
        )
        improper_dict = self.param_dict['improper']
        return self._get_from_choices(improper_dict, matching_orders)

    def get_nonbonded(self, key):
        """Get nonbonded parameters for a given atom type."""
        nonbonded_param = self.param_dict['nonbonded'][key]
        return nonbonded_param

    def get_from_topo_element(self, topo_element):
        """Get the parameter for a given topology element"""
        if isinstance(topo_element, Entities.Bond):
            return self.get_bond(topo_element.atom_types)
        elif isinstance(topo_element, Entities.Angle):
            return self.get_angle(topo_element.atom_types)
        elif isinstance(topo_element, Entities.Dihedral):
            return self.get_dihedral(topo_element.atom_types)
        elif isinstance(topo_element, Entities.Improper):
            return self.get_improper(topo_element.atom_types)
        else:
            raise ValueError('Invalid topology element type')

    def _apply_to_element_list(self, topo_type, topo_element_list):
        """Apply the parameter for a list of topology element"""
        if topo_type == 'bonds':
            param_get_func = self.get_bond
        elif topo_type == 'angles':
            param_get_func = self.get_angle
        elif topo_type == 'dihedrals':
            param_get_func = self.get_dihedral
        elif topo_type == 'impropers':
            param_get_func = self.get_improper
        else:
            raise ValueError('Invalid topology element type')
        no_param_list = []
        for topo_element in topo_element_list:
            param = param_get_func(topo_element.atom_types)
            if param is None:
                no_param_list.append(topo_element)
            else:
                topo_element.param = param
        return no_param_list
    
    def apply(self, topo_element_container: TopologyElementContainer):
        """Apply the parameter for a list of topology element"""
        if not isinstance(topo_element_container, TopologyElementContainer):
            raise TypeError(
                'Invalid argument type provided! TopologyElementContainer'
                f' is expected. {type(topo_element_container)} is provided.'
            )
        missing_param_dict = {}
        
        for topo_type, topo_element_list in topo_element_container:
            if topo_element_list is None:
                warnings.warn(
                    f'No {topo_type} found in '
                    f'{topo_element_container.containing_entity}.')
                continue
            no_param_list = self._apply_to_element_list(
                topo_type, topo_element_list
            )
            if no_param_list:
                warnings.warn(
                    f'{len(no_param_list)} {topo_type} failed to find '
                    'parameters.'
                )
                missing_param_dict[topo_type] = no_param_list
        topo_element_container.missing_param_dict = missing_param_dict

    @staticmethod
    def _find_atom_type(atom_name_list, residue_definition):
        """Find the atom type for a given atom name list."""
        atom_type_list = []
        for atom in atom_name_list:
            if atom in residue_definition:
                atom_type = residue_definition[atom].atom_type
            elif atom in residue_definition.removed_atom_dict:
                atom_type = residue_definition.removed_atom_dict[atom].atom_type
            elif atom == 'BLNK':
                ## TODO: use logging to log this
                return None # ic containing BLNK atom should be ignored
            else:
                raise ValueError(
                    f'Atom {atom} not found in residue definition: {residue_definition}'
                )
            atom_type_list.append(atom_type)
        return atom_type_list
    
    def res_def_fill_ic(self, residue_definition, preserve = True):
        """Fill in the missing parameters for the internal coordinates table
        of a residue definition."""
        for atom_key, ic_table in residue_definition.ic.items():
            atom_key = [atom.lstrip('+').lstrip('-') for atom in atom_key]
            atom_types = self._find_atom_type(atom_key, residue_definition)
            if atom_types is None:
                # ic containing BLNK atom should be ignored
                continue
            for ic_type in ic_table:
                if ic_type == 'Phi' or ic_type == 'improper':
                    continue
                if (ic_table[ic_type] is not None) and preserve:
                    continue
                ids = self.ic_position_dict[ic_type]
                cur_ic_atom_types = tuple(atom_types[i] for i in ids)
                ic_table[ic_type] = self._find_ic_param(cur_ic_atom_types)

    def _find_ic_param(self, key):
        if len(key) == 2:
            bond_param = self.get_bond(key)
            return bond_param.b0
        else:
            angle_param = self.get_angle(key)
            return angle_param.theta0

    def fill_ic(self, topology_loader, preserve = True):
        """Fill in the missing parameters for the internal coordinates table
        of a topology."""
        for residue_definition in topology_loader.residues:
            self.res_def_fill_ic(residue_definition, preserve)
        # Also fill in the patched residue definitions
        for residue_definition in topology_loader.patched_defs.values():
            self.res_def_fill_ic(residue_definition, preserve)


import numpy as np
from scipy.optimize import minimize

def energy(coord, atoms, params):
    """
    Computes the total energy of the protein structure using the given coordinates, 
    atoms, and force field parameters.
    """
    # Compute pairwise distances between atoms
    n_atoms = len(atoms)
    dist = np.zeros((n_atoms, n_atoms))
    for i in range(n_atoms):
        for j in range(i + 1, n_atoms):
            diff = coord[i] - coord[j]
            dist[i, j] = np.sqrt(np.sum(diff ** 2))
            dist[j, i] = dist[i, j]

    # Compute energy contributions from bonds, angles, and dihedrals
    bond_energy = 0.0
    angle_energy = 0.0
    dihedral_energy = 0.0
    for i in range(n_atoms):
        for j in range(i + 1, n_atoms):
            r = dist[i, j]
            sigma = (params[atoms[i]][0] + params[atoms[j]][0]) / 2.0
            epsilon = np.sqrt(params[atoms[i]][1] * params[atoms[j]][1])
            bond_energy += params[atoms[i]][2] * (r - sigma) ** 2
            angle_energy += params[atoms[i]][3] * (1.0 - np.cos(params[atoms[i]][4] * (r - sigma)))
            for k in range(j + 1, n_atoms):
                s = dist[j, k]
                phi = np.arccos(np.dot(coord[i] - coord[j], coord[k] - coord[j]) / (r * s))
                dihedral_energy += params[atoms[i]][5] * (1.0 + np.cos(params[atoms[i]][6] * phi))

    # Compute energy contribution from non-bonded interactions
    nb_energy = 0.0
    for i in range(n_atoms):
        for j in range(i + 1, n_atoms):
            r = dist[i, j]
            sigma = (params[atoms[i]][0] + params[atoms[j]][0]) / 2.0
            epsilon = np.sqrt(params[atoms[i]][1] * params[atoms[j]][1])
            nb_energy += 4.0 * epsilon * ((sigma / r) ** 12 - (sigma / r) ** 6)

    # Compute total energy
    total_energy = bond_energy + angle_energy + dihedral_energy + nb_energy
    return total_energy

def compute_energy_gradient(coords, bonds, angles, atom_types, ff_params):
    """
    Compute the energy gradient of a protein structure.

    Parameters:
    coords (np.ndarray): Atomic coordinates of the structure in angstroms.
    bonds (np.ndarray)
    angles (np.ndarray)
    atom_types (np.ndarray): Array of atomic types, where each element is an integer representing the atomic number.
    ff_params (dict): Dictionary of force field parameters, containing the force constants and Lennard-Jones parameters.

    Returns:
    np.ndarray: The energy gradient vector.
    """

    # Define constants
    N = coords.shape[0]  # number of atoms
    k_bond = ff_params['k_bond']  # bond force constant
    r_bond = ff_params['r_bond']  # bond dist constant
    k_angle = ff_params['k_angle']  # angle force constant
    theta_0 = ff_params['theta_0']  # reference angle 
    eps_lj = ff_params['eps_lj']  # Lennard-Jones epsilon parameter
    sig_lj = ff_params['sig_lj']  # Lennard-Jones sigma parameter

    # Initialize the energy gradient vector to zero
    energy_grad = np.zeros_like(coords)

    # Compute the bond contribution to the energy gradient
    for bond in bonds:
        i, j = bond
        r_ij = coords[j] - coords[i]
        dist = np.linalg.norm(r_ij)
        r_ij_hat = r_ij / dist
        force = -2 * k_bond[bond] * (dist - r_bond[bond]) * r_ij_hat
        energy_grad[i] += force
        energy_grad[j] -= force

    # Compute the angle contribution to the energy gradient
    for angle in angles:
        i, j, k = angle
        r_ij = coords[j] - coords[i]
        r_jk = coords[k] - coords[j]
        dist_ij = np.linalg.norm(r_ij)
        dist_jk = np.linalg.norm(r_jk)
        r_ij_hat = r_ij / dist_ij
        r_jk_hat = r_jk / dist_jk
        cos_theta = np.dot(r_ij_hat, r_jk_hat)
        sin_theta = np.sqrt(1 - cos_theta**2)
        force_ij = -2 * k_angle[angle] * (cos_theta - np.cos(theta_0[angle]))\
              / dist_ij / sin_theta * (r_jk_hat - cos_theta * r_ij_hat)
        force_jk = -2 * k_angle[angle] * (cos_theta - np.cos(theta_0[angle]))\
              / dist_jk / sin_theta * (r_ij_hat - cos_theta * r_jk_hat)
        energy_grad[i] += force_ij
        energy_grad[j] -= force_ij + force_jk
        energy_grad[k] += force_jk

    # Compute the Lennard-Jones contribution to the energy gradient
    for i in range(N):
        for j in range(i + 1, N):
            r_ij = coords[j] - coords[i]
            dist = np.linalg.norm(r_ij)
            r_ij_hat = r_ij / dist
            eps_ij = np.sqrt(eps_lj[atom_types[i]] * eps_lj[atom_types[j]])
            sig_ij = 0.5 * (sig_lj[atom_types[i]] + sig_lj[atom_types[j]])
            factor = 24 * eps_ij * (2 * (sig_ij / dist)**12 - (sig_ij / dist)**6) / dist**2
            energy_grad[i] += factor * r_ij_hat
            energy_grad[j] -= factor * r_ij_hat

    return energy_grad


import os
import warnings
import pickle
from typing import List, Tuple
from copy import deepcopy
from Bio.Data.PDBData import protein_letters_3to1_extended
from Bio.Data.PDBData import protein_letters_1to3
from crimm import StructEntities as Entities
from crimm.IO.RTFParser import RTFParser
from crimm.Modeller import ResidueFixer

toppar_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../Data/toppar')
)
rtf_path_dict = {
    "protein": os.path.join(toppar_dir, 'prot.rtf'),
    "nucleic": os.path.join(toppar_dir, 'na.rtf'),
    "lipid": os.path.join(toppar_dir, 'lipid.rtf'),
    "carb": os.path.join(toppar_dir, 'carb.rtf'),
    "ethers": os.path.join(toppar_dir, 'ethers.rtf'),
    "cgenff": os.path.join(toppar_dir, 'cgenff.rtf')
}


def _find_atom_in_residue(
        residue: Entities.Residue, atom_name: str
    )->Entities.Atom:
    if atom_name.startswith('+') or atom_name.startswith('-'):
        return None
    if atom_name in residue:
        return residue[atom_name]
    if atom_name.startswith('H'):
        return residue.missing_hydrogens[atom_name]
    return residue.missing_atoms[atom_name]

def _find_atom_from_neighbor(
        cur_residue: Entities.Residue, atom_name: str
    )->Entities.Atom:
    """Get atom from neighbor residue. (Private function)"""
    resseq = cur_residue.id[1]
    chain = cur_residue.parent
    if atom_name.startswith('-') and resseq-1 in chain:
        atom_name = atom_name[1:]
        neighbor_residue = chain[resseq-1]
        return _find_atom_in_residue(neighbor_residue, atom_name)
    elif atom_name.startswith('+') and resseq+1 in chain:
        atom_name = atom_name[1:]
        neighbor_residue = chain[resseq+1]
        return _find_atom_in_residue(neighbor_residue, atom_name)
    return None

def get_bonds_within_residue(residue: Entities.Residue)->List[Entities.Bond]:
    """Return a list of bonds within the residue (peptide bonds linking neighbor 
    residues are excluded). Raise ValueError if the topology definition is 
    not loaded."""
    if residue.topo_definition is None:
        raise ValueError(
            'Topology definition is not loaded for this residue!'
        )
    bonds = []
    bond_dict = residue.topo_definition.bonds
    for bond_type, bond_list in bond_dict.items():
        for atom_name1, atom_name2 in bond_list:
            atom1 = _find_atom_in_residue(residue, atom_name1)
            atom2 = _find_atom_in_residue(residue, atom_name2)
            if atom1 is None or atom2 is None:
                # one of the atoms has to be in the neighbor residue
                continue
            bonds.append(
                Entities.Bond(atom1, atom2, bond_type)
            )
    return bonds

def atom_add_neighbors(atom1: Entities.Atom, atom2: Entities.Atom):
    """Add atom2 to atom1's neighbors and vice versa"""
    atom1.neighbors.add(atom2)
    atom2.neighbors.add(atom1)

def residue_trace_atom_neigbors(residue: Entities.Residue)->List[Entities.Bond]:
    """Trace all bonds within the residue and add the atoms to each other's
    neighbors list. Return a list of bonds within the residue."""
    bonds = get_bonds_within_residue(residue)
    for bond in bonds:
        a1, a2 = bond
        atom_add_neighbors(a1, a2)
    return bonds

def chain_clear_atom_neighbors(chain: Entities.PolymerChain):
    """Clear all neighbors of atoms in the chain"""
    for atom in chain.get_atoms():
        atom.neighbors = set()

def chain_trace_atom_neighbors(
        chain: Entities.PolymerChain, inter_res_bonding_atoms: Tuple[str],
        bond_type='single'
    )->List[Entities.Bond]:
    """Trace all bonds within the chain and add the atoms to each other's
    neighbors list. Return a list of bonds within the chain."""
    end_atom, start_atom = inter_res_bonding_atoms
    # e.g. this would be C-N for peptide,
    # thus end_atom is C and start_atom is N
    chain.sort_residues()
    # Clearing neighbors is needed for patching and regenerating topology
    chain_clear_atom_neighbors(chain)
    all_bonds = []
    for i, cur_res in enumerate(chain.residues[:-1]):
        next_res = chain.residues[i+1]
        bonds = residue_trace_atom_neigbors(cur_res)
        all_bonds.extend(bonds)
        # add inter-residue neigbors from peptide/nucleotide bond

        a1 = _find_atom_in_residue(cur_res, end_atom)
        a2 = _find_atom_in_residue(next_res, start_atom)
        atom_add_neighbors(a1, a2)
        inter_res_bond = Entities.Bond(a1, a2, bond_type)
        all_bonds.append(inter_res_bond)

    last_res = chain.residues[-1]
    bonds = residue_trace_atom_neigbors(last_res)
    all_bonds.extend(bonds)
    return all_bonds

def _add_dihedral(cur_atom, nei_atom, second_nei_atom, dihedral_set):
    """Add dihedral angle to the dihedral set. (Private function)"""
    for third_nei_atom in second_nei_atom.neighbors:
        if third_nei_atom == nei_atom:
            continue
        dihe = Entities.Dihedral(cur_atom, nei_atom, second_nei_atom, third_nei_atom)
        dihedral_set.add(dihe)

def _add_angle_and_dihedral(cur_atom, nei_atom, angle_set, dihedral_set):
    """Add angle and dihedral angle to the angle and dihedral sets. (Private function)"""
    for second_nei_atom in nei_atom.neighbors:
        if second_nei_atom == cur_atom:
            continue
        angle = Entities.Angle(cur_atom, nei_atom, second_nei_atom)
        angle_set.add(angle)
        _add_dihedral(cur_atom, nei_atom, second_nei_atom, dihedral_set)

def traverse_graph(cur_atom, angle_set, dihedral_set, visited_set):
    """Traverse the graph of atoms and add all angles and dihedral angles to the
    angle and dihedral sets."""
    visited_set.add(cur_atom)

    for nei_atom in cur_atom.neighbors:
        _add_angle_and_dihedral(cur_atom, nei_atom, angle_set, dihedral_set)
        if nei_atom in visited_set:
            continue
        traverse_graph(nei_atom, angle_set, dihedral_set, visited_set)

def _get_improper_from_atom_names(
        residue: Entities.Residue, atom_names: Tuple[str]
    )->Entities.Improper:
    """Get improper from atom names. (Private function)"""
    atoms = []
    for atom_name in atom_names:
        if atom_name.startswith('-') or atom_name.startswith('+'):
            atom = _find_atom_from_neighbor(residue, atom_name)
        else:
            atom = _find_atom_in_residue(residue, atom_name)
        if atom is None:
            return None
        atoms.append(atom)
    a1, a2, a3, a4 = atoms
    for atom in (a2, a3, a4):
        if atom not in a1.neighbors:
            raise ValueError(
                'Improper angle definition is incorrect: '
                f'Atom {atom} is not neighbor of atom {a1}!'
            )
    return Entities.Improper(a1, a2, a3, a4)

def _is_terminal_or_orphan_residue(residue: Entities.Residue)->bool:
    """Check if the residue is a terminal residue or does not belong to a chain. 
    (Private function)"""
    chain: Entities.PolymerChain = residue.parent
    if chain is None:
        # orphan residue
        return True
    return residue in (chain.residues[0], chain.residues[-1])

def residue_get_impropers(residue: Entities.Residue)->List[Entities.Improper]:
    """Return a list of improper angles within the residue. Raise ValueError if the 
    topology definition is not loaded."""
    if residue.topo_definition is None:
        raise ValueError(
            'Topology definition is not loaded for this residue!'
        )
    impropers = []
    for impr_atom_names in residue.topo_definition.impropers:
        improper = _get_improper_from_atom_names(residue, impr_atom_names)
        if improper is None:
            if not _is_terminal_or_orphan_residue(residue):
                warnings.warn(
                    f'Cannot find improper {impr_atom_names} in residue {residue}'
                )
            continue
        impropers.append(improper)
    return impropers

def get_impropers(chain: Entities.PolymerChain)->List[Entities.Improper]:
    """Return a list of improper angles within the chain. Raise ValueError if the 
    topology definition is not loaded."""
    impropers = []
    for res in chain.residues:
        impropers.extend(residue_get_impropers(res))
    return impropers

def _get_cmap_from_atom_names(res, cmap_atom_names):
    """Get cmap from atom names. (Private function)"""
    raise NotImplementedError

def get_cmap(chain: Entities.PolymerChain)->List[Entities.CMap]:
    """Return a list of CMap terms within the chain. Raise ValueError if the 
    topology definition is not loaded."""
    cmaps = []
    for res in chain.residues:
        if res.topo_definition is None:
            raise ValueError(
                'Topology definition is not loaded for this residue!'
            )
        for cmap_atom_names in res.topo_definition.cmaps:
            cmap = _get_cmap_from_atom_names(res, cmap_atom_names)
            if cmap is None:
                if not _is_terminal_or_orphan_residue(res):
                    warnings.warn(
                        f'Cannot find cmap {cmap_atom_names} in residue {res}'
                    )
                continue
            cmaps.append(cmap)
    return cmaps

##TODO: separate this class into two classes: TopologyLoader and TopologyGenerator
class TopologyLoader:
    """Class for loading topology definition to the residue and find any missing atoms."""
    def __init__(self, entity_type=None, data_dict_path=None):
        self.rtf_version = None
        self.res_defs = {}
        self.residues = []
        self.patches = []
        self.patched_defs = {}
        self._raw_data_strings = []

        if data_dict_path is not None:
            with open(data_dict_path, 'rb') as f:
                data_dict = pickle.load(f)
                self.load_data_dict(data_dict)

        elif entity_type is not None:
            entity_type = entity_type.lower()
            if entity_type not in rtf_path_dict:
                raise ValueError(f'Unknown entity type: {entity_type}')
            rtf = RTFParser(file_path=rtf_path_dict[entity_type])
            self._raw_data_strings = rtf.lines
            self.load_data_dict(rtf.topo_dict, rtf.rtf_version)

    def load_data_dict(self, topo_data_dict: dict, rtf_version:str=None):
        """Load topology data from a dictionary. The dictionary should be parsed
        from a RTF file."""
        self.rtf_version = rtf_version
        for resname, res_topo_dict in topo_data_dict.items():
            # if resname in Entities.ResidueDefinition.na_3to1:
            #     # Map 3-letter residue name to 1-letter residue name for nucleic
            #     # acids, since biopython uses 1-letter residue name for them.
            #     resname = Entities.ResidueDefinition.na_3to1[resname]

            if res_topo_dict['is_patch']:
                res_def = Entities.PatchDefinition(
                    self.rtf_version, resname, res_topo_dict
                )
                self.patches.append(res_def)
            else:
                res_def = Entities.ResidueDefinition(
                    self.rtf_version, resname, res_topo_dict
                )
                self.residues.append(res_def)

            self.res_defs[resname] = res_def

        if 'HIS' not in self.res_defs and 'HSD' in self.res_defs:
            # Map all histidines HIS to HSD
            self.res_defs['HIS'] = self.res_defs['HSD']

    def __repr__(self):
        return (
            f'<TopologyLoader Ver={self.rtf_version} '
            f'Contains {len(self.residues)} RESIDUE and '
            f'{len(self.patches)} PATCH definitions>'
        )

    def __getitem__(self, __key: 'str'):
        if __key in Entities.ResidueDefinition.na_1to3:
            __key = Entities.ResidueDefinition.na_1to3[__key]
        return self.res_defs[__key]
    
    def __contains__(self, __key: 'str'):
        if __key in Entities.ResidueDefinition.na_1to3:
            __key = Entities.ResidueDefinition.na_1to3[__key]
        return __key in self.res_defs
    
    def __iter__(self):
        return iter(self.res_defs.values())
    
    def generate_residue_topology(
            self, residue: Entities.Residue, coerce = False, QUIET = False
        ):
        """Load topology definition into the residue and find any missing atoms.
        Argument:
            residue: the Residue object whose topology is to be defined
            coerce: if True, try to coerce the modified residue name to the canonical name
                    and load the canonical residue topology definition
            QUIET: if True, suppress all warnings

        Return:
            True if the residue is defined, False otherwise"""

        if isinstance(residue, Entities.DisorderedResidue):
            return self.generate_residue_topology(
                residue.selected_child, coerce=coerce, QUIET=QUIET
            )

        if residue.resname not in self:
            if not QUIET:
                warnings.warn(
                    f"Residue {residue.resname} is not defined in the topology file!"
                )
            if residue.topo_definition is not None:
                return True
            if not coerce:
                return False
            success = self.coerce_resname(residue, QUIET=QUIET)
            if not success:
                return False

        if residue.topo_definition is not None and not QUIET:
            warnings.warn("Topology definition already exists! Overwriting...")

        res_definition = self[residue.resname]
        self.apply_topo_def_on_residue(residue, res_definition, QUIET=QUIET)
        return True

    @staticmethod
    def apply_topo_def_on_residue(
            residue: Entities.Residue,
            res_definition: Entities.ResidueDefinition,
            QUIET = False
        ):
        """Apply the topology definition to the residue"""
        residue.topo_definition = res_definition
        residue.total_charge = res_definition.total_charge
        residue.impropers = res_definition.impropers
        residue.cmap = res_definition.cmap
        residue.H_donors = res_definition.H_donors
        residue.H_acceptors = res_definition.H_acceptors
        residue.param_desc = res_definition.desc
        TopologyLoader._load_atom_groups(residue)
        residue.undefined_atoms = []
        for atom in residue:
            if atom.name not in res_definition:
                residue.undefined_atoms.append(atom)
                if not QUIET:
                    parent_id = (atom.parent.id[1], atom.parent.resname)
                    warnings.warn(
                        f"Atom {atom.name} from {parent_id} is not defined in "
                        "the topology file!"
                    )

    @staticmethod
    def _create_missing_atom(residue: Entities.Residue, atom_name: str)-> list:
        """Create and separate missing heavy atoms and missing hydrogen atom by atom name"""
        missing_atom : Entities.Atom = residue.topo_definition[atom_name].create_new_atom()
        missing_atom.set_parent(residue)
        if atom_name.startswith('H'):
            residue.missing_hydrogens[atom_name] = missing_atom
        else:
            residue.missing_atoms[atom_name] = missing_atom
        return missing_atom

    @staticmethod
    def _load_group_atom_topo_definition(
            residue: Entities.Residue, atom_name_list: list
        ) -> tuple:
        """Load topology definition to each atom in the residue and find any missing 
        atoms. 
        Argument:
            atom_name_list: list of atom names that are in the same group
        Return:
            atom_group: the Atom object in the group
        """
        atom_group = []
        for atom_name in atom_name_list:
            if atom_name not in residue:
                cur_atom = TopologyLoader._create_missing_atom(residue, atom_name)
            else:
                cur_atom = residue[atom_name]
                cur_atom.topo_definition = residue.topo_definition[atom_name]
            atom_group.append(cur_atom)
        return tuple(atom_group)

    @staticmethod
    def _load_atom_groups(residue: Entities.Residue):
        residue.atom_groups = []
        residue.missing_atoms, residue.missing_hydrogens = {},{}
        atom_group_lists = residue.topo_definition.atom_groups
        for atom_names in atom_group_lists:
            cur_group = TopologyLoader._load_group_atom_topo_definition(
                residue, atom_names
            )
            residue.atom_groups.append(cur_group)

    def coerce_resname(self, residue: Entities.Residue, QUIET = False)->bool:
        """Coerce the name of modified residue to reconstruct it as the canonical 
        one that it is based on. 

        Argument:
            residue: the residue whose name is to be coerced
        Return:
            True if the residue is a known modified residue and is successfully coerced
            False otherwise"""
        ## TODO: add examples to the docstring
        if residue.resname not in protein_letters_3to1_extended:
            return False
        # if the residue is a known modified residue,
        # coerce the residue name to reconstruct it as the
        # canonical one
        code = protein_letters_3to1_extended[residue.resname]
        new_resname = protein_letters_1to3[code]
        if not QUIET:
            warnings.warn(
                f'Coerced Residue {residue.resname} to {new_resname}'
            )
        residue.resname = new_resname
        _, resseq, icode = residue.id
        residue.id = (" ", resseq, icode)

        if residue.parent is not None:
            residue.parent.het_res.remove(residue)
            residue.parent.het_resseq_lookup.pop(resseq)

        if hasattr(residue.parent, "reported_res"):
            # if the chain has reported_res attribute, update it
            # to avoid generating new gaps due to mismatch resnames
            residue.parent.reported_res[resseq-1] = (resseq, new_resname)
        return True

    def generate_chain_topology(
            self, chain: Entities.Chain, coerce: bool = False,
            first_patch: str = None, last_patch: str = None,
            QUIET = False
        ):
        """Load topology definition into the chain and find any missing atoms.
        Argument:
            chain: the Chain object whose topology is to be defined
            coerce: if True, try to coerce the modified residue name to the canonical name
                    and load the canonical residue topology definition
            QUIET: if True, suppress all warnings
        """
        chain.undefined_res = []
        for residue in chain:
            is_defined = self.generate_residue_topology(
                residue, coerce=coerce, QUIET=QUIET
            )
            if not is_defined:
                chain.undefined_res.append(residue)
        if (n_undefined:=len(chain.undefined_res)) > 0 and not QUIET:
            warnings.warn(
                f"{n_undefined} residue(s) not defined in the chain!"
            )
        if first_patch is not None or last_patch is not None:
            self.patch_termini(chain, first_patch, last_patch, QUIET=True)

        topo_elements = TopologyElementContainer()
        chain.topo_elements = topo_elements.load_chain(chain)

    def patch_termini(
            self, chain: Entities.PolymerChain,
            first: str, last: str, QUIET=False
        ):
        """Patch the terminal residues of the chain"""
        if chain.chain_type not in ('Polypeptide(L)', 'Polyribonucleotide'):
            raise NotImplementedError(
                "Only polypeptide and polynucleotide chains are supported "
                f"for patching! Got {chain.chain_type}"
            )
        chain.sort_residues()
        if first is not None:
            self.patch_residue(chain.child_list[0], first, "NTER", QUIET=QUIET)
        if last is not None:
            self.patch_residue(chain.child_list[-1], last, "CTER", QUIET=QUIET)
        if chain.topo_elements is not None:
            # Update topology elements if they are already definedß
            topo_elements = TopologyElementContainer()
            chain.topo_elements = topo_elements.load_chain(chain)

    def patch_residue(
            self, residue: Entities.Residue, patch: str,
            patch_loc = 'MIDCHAIN', QUIET = False
        ):
        """Patch the residue with the patch definition"""
        if residue.topo_definition is None:
            raise ValueError(
                f"Cannot patch the first residue {residue.resname} "
                "because it is undefined (no topology definition exists)!"
            )
        res_def = residue.topo_definition
        patched_def_name = res_def.resname + "_" + patch
        if patched_def_name in self.patched_defs:
            patched_res_def = self.patched_defs[patched_def_name]
        else:
            patcher = ResiduePatcher()
            patch_def = self[patch]
            patched_res_def = patcher.patch_residue_definition(
                res_def, patch_def, patch_loc=patch_loc
            )
            self.patched_defs[patched_def_name] = patched_res_def

        self.apply_topo_def_on_residue(residue, patched_res_def, QUIET=QUIET)
        fixer = ResidueFixer()
        fixer.load_residue(residue)
        fixer.remove_undefined_atoms()


class TopologyElementContainer:
    """A class object that stores topology elements"""
    def __init__(self):
        self._visited_atoms = None
        self.bonds = None
        self.angles = None
        self.dihedrals = None
        self.impropers = None
        self.cmap = None
        self.atom_lookup = None
        self.missing_param_dict = None
        self.containing_entity = None

    def __iter__(self):
        topo_attrs = [
            'bonds', 'angles', 'dihedrals', 'impropers', 'cmap'
        ]
        for attr in topo_attrs:
            yield attr, getattr(self, attr)

    def __repr__(self) -> str:
        if self.containing_entity is None:
            return "<EmptyTopologyElementContainer>"
        s = f"<TopologyElementContainer for {self.containing_entity} with "
        for attr, value in self:
            if value is None:
                n = 0
            else:
                n = len(value)
            s += f"{attr}={n}, "
        s = s[:-2] + ">"
        return s

    def load_chain(self, chain: Entities.PolymerChain):
        """Find all topology elements from the chain"""
        self.containing_entity = chain
        self.find_topo_elements(chain)
        self.create_atom_lookup_table()
        return self
    
    @staticmethod
    def _find_seeding_atom(chain):
        """Find the first atom to start the search"""
        for atom in chain.child_list[0]:
            if len(atom.neighbors) > 0:
                return atom

    ## TODO: get Cmap from the topology rtf file
    def find_topo_elements(self, chain: Entities.Chain):
        """Find all topology elements in the chain"""
        if chain.chain_type == 'Polypeptide(L)':
            inter_res_bond = ('C','N') # peptide bond
        elif chain.chain_type == 'Polyribonucleotide':
            inter_res_bond = ("O3'",'P') # phosphodiester bond
        else:
            raise NotImplementedError("Chain type not supported yet!")
        
        if chain.undefined_res:
            raise ValueError(
                "Cannot find topology elements for a chain with undefined residues! "
                f"Undefined residues: {chain.undefined_res}. This is possibly due "
                "to heterogen residues. Use Chain.generate_chain_topology() "
                "with coerce=True to coerce the heterogen into canoncal residue "
                "if necessary."
            )

        if not chain.is_continuous():
            raise ValueError(
                "Cannot generate topology elements for a chain with "
                f"discontinuous backbone! Missing residues: {chain.gaps}. Use "
                "loop modeling tool to construct the missing residues first."
            )

        self.bonds = chain_trace_atom_neighbors(chain, inter_res_bond)
        visited_atoms = set()
        angles = set()
        dihedrals = set()
        seed_atom = self._find_seeding_atom(chain)
        traverse_graph(seed_atom, angles, dihedrals, visited_atoms)
        self._visited_atoms = list(visited_atoms)
        self.angles = list(angles)
        self.dihedrals = list(dihedrals)
        self.impropers = get_impropers(chain)

    def create_atom_lookup_table(self) -> dict:
        """Create a lookup table for all topology elements for a given atom in the chain"""
        atom_lookup = {
            atom:{
                'bonds': [],
                'angles': [],
                'dihedrals': [],
                'impropers': []
            }
            for atom in self._visited_atoms
        }

        for bond in self.bonds:
            for atom in bond:
                atom_lookup[atom]['bonds'].append(bond)
            
        for angle in self.angles:
            for atom in angle:
                atom_lookup[atom]['angles'].append(angle)

        for dihedral in self.dihedrals:
            for atom in dihedral:
                atom_lookup[atom]['dihedrals'].append(dihedral)

        for improper in self.impropers:
            for atom in improper:
                atom_lookup[atom]['impropers'].append(improper)

        self.atom_lookup = atom_lookup

##TODO: Rewrite TopoDef and this class to use the TopologyElementContainer instead
## of removing entries one by one here
class ResiduePatcher:
    """Class Object for patching a residue with a patch definition"""
    def __init__(self):
        self.res: Entities.ResidueDefinition = None
        self.patch: Entities.PatchDefinition = None
        # for marking the ic entry to be removed from terminal residue during 
        # patching, either '+' or '-' depending on the patch location
        self.remove_nei_ic_prefix: str = None

    def _remove_atom_from_bonds(self, atom_name:str):
        """Remove the atom from the bonds"""
        for bonds in self.res.bonds.values():
            remove_list = []
            for bond in bonds:
                if atom_name in bond:
                    remove_list.append(bond)
            for bond in remove_list:
                bonds.remove(bond)
        
    def _remove_neighbor_atom_from_ic(self):
        remove_keys = set()
        for ic_key in self.res.ic:
            for atom in ic_key:
                if atom.startswith(self.remove_nei_ic_prefix):
                    remove_keys.add(ic_key)
        for ic_key in remove_keys:
            self.res.ic.pop(ic_key)

    def _remove_neighbor_atom_from_cmap(self):
        remove_keys = set()
        for cmap in self.res.cmap:
            for dihe in cmap:
                for atom in dihe:
                    if atom.startswith(self.remove_nei_ic_prefix):
                        remove_keys.add(cmap)

        for cmap in remove_keys:
            self.res.cmap.remove(cmap)

    def _remove_neighbor_atom_from_improper(self):
        remove_keys = set()
        for improper in self.res.impropers:
            for atom in improper:
                if atom.startswith(self.remove_nei_ic_prefix):
                    remove_keys.add(improper)

        for improper in remove_keys:
            self.res.impropers.remove(improper)

    def _remove_atom_from_ic(self, atom_name:str):
        """Remove the atom from the ic table dictionary"""
        remove_keys = []
        for ic_atom_names in self.res.ic:
            if atom_name in ic_atom_names:
                remove_keys.append(ic_atom_names)
        for key in remove_keys:
            self.res.ic.pop(key)

    def _remove_atom_from_param(self, param_attr, atom_name:str):
        """Remove the residue from the parameter attribute of the residue"""
        remove_keys = set()
        for iterable in param_attr:
            if atom_name in iterable:
                remove_keys.add(iterable)
        
        for iterable in remove_keys:
            param_attr.remove(iterable)

    def _remove_atom_from_cmap(self, cmaps, atom_name:str):
        """Remove the residue from the parameter attribute of the residue"""
        remove_keys = set()
        for cmap in cmaps:
            for dihe in cmap:
                if atom_name in dihe:
                    remove_keys.add(cmap)

        for cmap in remove_keys:
            cmaps.remove(cmap)

    def _delete_atom_params(self):
        """Delete the parameters of the atoms that are deleted in the patch"""
        if self.patch.delete is None:
            return
        delete_atom_names = []
        for entity_type, entity_name in self.patch.delete:
            if entity_type == 'ATOM' and entity_name in self.res:
                delete_atom_names.append(entity_name)
        for atom_name in delete_atom_names:
            self.res.removed_atom_dict[atom_name] = self.res.atom_dict.pop(atom_name)
            self._remove_atom_from_bonds(atom_name)
            self._remove_atom_from_ic(atom_name)
            for param_attr in (
                self.res.impropers, self.res.H_donors,
                self.res.H_acceptors, self.res.atom_groups
            ):
                self._remove_atom_from_param(param_attr, atom_name)
            self._remove_atom_from_cmap(self.res.cmap, atom_name)

    def _apply_patch(self):
        """Apply the patch on the residue definition"""
        self._delete_atom_params()
        self.res.atom_dict.update(self.patch.atom_dict)
        for bond_type in self.res.bonds:
            self.res.bonds[bond_type].extend(self.patch.bonds[bond_type])

        self.res.total_charge = 0
        for atom_def in self.res:
            self.res.total_charge += atom_def.charge
        self.res.atom_groups.extend(self.patch.atom_groups)
        self.res.H_donors.extend(self.patch.H_donors)
        self.res.H_acceptors.extend(self.patch.H_acceptors)
        self.res.impropers.extend(self.patch.impropers)
        self.res.cmap.extend(self.patch.cmap)

        self.res.ic = {**self.res.ic, **self.patch.ic}
        if self.remove_nei_ic_prefix is not None:
            self._remove_neighbor_atom_from_ic()
            self._remove_neighbor_atom_from_improper()
            self._remove_neighbor_atom_from_cmap()

        self.res.is_modified = True

    def patch_residue_definition(
            self,
            residue_definition: Entities.ResidueDefinition,
            patch_definition: Entities.PatchDefinition,
            patch_loc: str = "MIDCHAIN",
        ) -> Entities.ResidueDefinition:
        """Patch a residue definition with a patch definition. Return the patched residue definition
        """
        patch_loc = patch_loc.upper()
        if patch_loc not in ("MIDCHAIN", "NTER", "CTER"):
            raise ValueError("Patch type must be 'MIDCHAIN', 'NTER' or 'CTER'")
        if patch_loc == "NTER":
            self.remove_nei_ic_prefix = '-'
        elif patch_loc == "CTER":
            self.remove_nei_ic_prefix = '+'

        # need to make a copy of the patch definition
        # to avoid modifying the original patch definition
        self.res = deepcopy(residue_definition)
        # remove the standard residue coordinates
        self.res.standard_coord_dict = None
        self.res._standard_res = None
        self.patch = deepcopy(patch_definition)
        self._apply_patch()
        self.res.assign_donor_acceptor()
        self.res.create_atom_lookup_dict()
        self.res.patch_with = self.patch.resname
        return self.res


from crimm.Modeller.TopoFixer import ResidueFixer
from crimm.Modeller.TopoLoader import TopologyLoader
from crimm.Modeller.ParamLoader import ParameterLoader
from crimm.Modeller.SeqChainGenerator import SeqChainGenerator


import os
import io
import requests
import warnings
from Bio.Seq import Seq
from crimm.IO import MMCIFParser, PDBParser
from crimm.Superimpose.ChainSuperimposer import ChainSuperimposer

def uniprot_id_query(pdbid, entity_id):
    """Query wth the RCSB PDB API for uniprot id for a given pdb id and 
    chain entity id
    """
    rcsb_url = f"https://data.rcsb.org/rest/v1/core/uniprot/{pdbid}/{entity_id}"
    response = requests.get(rcsb_url, timeout=200)
    if (code := response.status_code) != 200:
        warnings.warn(
            f"GET request on RCSB for \"{pdbid}-{entity_id}\" for uniprot ID "
            f"did not return valid result! \n[Status Code] {code}"
        )
        return
    kw1 = 'rcsb_uniprot_container_identifiers'
    kw2 = 'uniprot_id'
    uniprot_id = response.json()[0][kw1][kw2]
    return uniprot_id

def _find_local_cif_path(pdb_id, entry_point):
    """Find the path to a local cif file"""
    pdb_id = pdb_id.lower()
    subdir = pdb_id[1:3]
    file_path = os.path.join(entry_point, subdir, pdb_id+'.cif')
    if os.path.exists(file_path):
        return file_path

def _file_handle_from_url(cif_url):
    """Get a cif file from a url, return a file handle to the cif file"""
    f_handle = io.StringIO()
    result = requests.get(cif_url,timeout=10)
    if (code := result.status_code) != 200:
        warnings.warn(
            "GET request for file did not return valid result!\n"
            f"[Status Code] {code}"
        )
        return
    f_handle.write(result.content.decode('utf-8'))
    f_handle.seek(0)
    return f_handle

def _get_alphafold_fh(uniprot_id):
    """Get a mmcif file handle from the alphafold database for a given uniprot id
    """
    query_url = f"https://alphafold.ebi.ac.uk/api/prediction/{uniprot_id}"
    response = requests.get(query_url, timeout=500)
    if (code := response.status_code) != 200:
        warnings.warn(
            f"GET request on AlphaFold DB for \"{uniprot_id}\"  did not return "
            f"valid result! \n[Status Code] {code}"
        )
        return
    alphafold_cif_url = response.json()[0]['cifUrl']
    return _file_handle_from_url(alphafold_cif_url)

def fetch_alphafold(uniprot_id):
    """Get a structure from the alphafold database for a given uniprot id"""
    msg = f"AlphaFold DB failed find structure for {uniprot_id}"
    if uniprot_id is None:
        # We only issue warning here
        warnings.warn(msg)
        return
    file = _get_alphafold_fh(uniprot_id)
    if file is None:
        # We only issue warning here
        warnings.warn(msg)
        return
    parser = MMCIFParser(
        # AlphaFold only has one model and does not have multiple 
        # assemblies, solvent, or hydrogens
        first_model_only = True,
        first_assembly_only = False, 
        include_hydrogens = False,
        include_solvent = False
    )
    structure = parser.get_structure(file)
    return structure

def _get_mmcif_fh(pdb_id):
    """Get a mmcif file handle from the rcsb database"""
    entry_point = "https://files.rcsb.org/download"
    rcsb_cif_url = f"{entry_point}/{pdb_id}.cif"
    return _file_handle_from_url(rcsb_cif_url)

def fetch_rcsb(
        pdb_id,
        local_entry = None,
        first_model_only = True,
        first_assembly_only = True,
        include_solvent = True,
        include_hydrogens = False,
    ):
    """Get a structure from rcsb with a pdb id or from a local mmcif file"""
    if local_entry is not None:
        file = _find_local_cif_path(pdb_id, entry_point = local_entry)
    else:
        file = _get_mmcif_fh(pdb_id)
    if file is None:
        raise ValueError(f"Could not load file for {pdb_id}")
    parser = MMCIFParser(
        first_model_only = first_model_only,
        first_assembly_only = first_assembly_only,
        include_hydrogens = include_hydrogens,
        include_solvent=include_solvent
    )
    structure = parser.get_structure(file)
    return structure

def fetch_swiss_model(uniprot_id):
    """Get the first matching stucuture from the Swiss Model database for a given 
    uniprot id
    """
    base_url = (
        "https://swissmodel.expasy.org/repository/uniprot/{uniprot_id}.{ext}"
        "?provider=swissmodel"
    )
    header_url = base_url.format(uniprot_id = uniprot_id, ext = 'json')
    struct_url = base_url.format(uniprot_id = uniprot_id, ext = 'pdb')
    result = requests.get(header_url,timeout=10)
    if (code := result.status_code) != 200:
        warnings.warn(
            "GET request for header did not return valid result!\n"
            f"[Status Code] {code}"
        )
        return
    info_dict = result.json()['result']
    can_seq_str = info_dict['sequence']
    chain_info_dict = info_dict['structures'][0]['chains'][0]
    desc = chain_info_dict['segments'][0]['smtl']['description']

    parser = PDBParser()
    struct_fh = _file_handle_from_url(struct_url)
    if struct_fh is None:
        return
    structure = parser.get_structure(struct_fh, f'{uniprot_id}-SwissModel')

    for chain in structure.child_list[0]:
        if not hasattr(chain, 'seq'):
            continue
        if str(chain.seq) in can_seq_str:
            chain.can_seq = Seq(can_seq_str)
            chain.pdbx_description = desc
    return structure

def fetch_swiss_model_multiple(uniprot_id):
    """Get all stucutures from the Swiss Model database for a given uniprot id
    """
    header_url = (
        f"https://swissmodel.expasy.org/repository/uniprot/{uniprot_id}.json"
        "?provider=swissmodel"
    )
    result = requests.get(header_url,timeout=10)
    if (code := result.status_code) != 200:
        warnings.warn(
            "GET request for header did not return valid result!\n"
            f"[Status Code] {code}"
        )
        return

    info_dict = result.json()['result']
    can_seq_str = info_dict['sequence']
    chain_info_dict = info_dict['structures'][0]['chains'][0]
    desc = chain_info_dict['segments'][0]['smtl']['description']

    struct_info_dicts = result.json()['result']['structures']
    structures = []
    for info_dict in struct_info_dicts:
        structures.append(
            _sm_struct_from_info_dict(uniprot_id, info_dict, can_seq_str, desc)
        )
    return structures

def _sm_struct_from_info_dict(uniprot_id, struct_info_dict, can_seq_str, desc):
    parser = PDBParser()
    structure_url = struct_info_dict['coordinates']
    struct_fh = _file_handle_from_url(structure_url)
    if struct_fh is None:
        return
    structure = parser.get_structure(struct_fh, f'{uniprot_id}-SwissModel')

    for chain in structure.models[0]:
        if not hasattr(chain, 'seq'):
            continue
        if str(chain.seq) in can_seq_str:
            chain.can_seq = Seq(can_seq_str)
            chain.pdbx_description = desc

    return structure

def _fetch_with_chain(chain, fetcher):
    if not hasattr(chain.parent, 'parent') or chain.parent.parent is None:
        raise ValueError("Chain has no parent structure!")
    pdb_id = chain.parent.parent.id
    if pdb_id is None or len(pdb_id) != 4:
        raise ValueError(f"Chain's PDB ID {pdb_id} is invalid!")
    entity_id = chain.entity_id
    uniprot_id = uniprot_id_query(pdb_id, entity_id)
    new_struct =  fetcher(uniprot_id)
    if new_struct is None:
        warnings.warn(
            f"Could not find AlphaFold structure for {pdb_id}-{entity_id}"
        )
        return
    new_chain = new_struct.models[0].chains[0]
    imposer = ChainSuperimposer()
    imposer.set_chains(chain, new_chain)
    imposer.apply_transform(new_struct)
    return new_struct

def fetch_alphafold_from_chain(chain):
    """Find the alphafold structure for a given chain. The chain must have a parent 
    structure with a valid PDB ID. The chain must also have an entity id assigned 
    from mmCIF.
    The returned AlphaFold structure will have only one model and one chain, and the
    chain will be superimposed to the input chain.
    """
    return _fetch_with_chain(chain, fetch_alphafold)

def fetch_swiss_model_from_chain(chain):
    """Find the Swiss Model structure for a given chain. The chain must have a parent 
    structure with a valid PDB ID. The chain must also have an entity id assigned 
    from mmCIF.
    The returned Swiss Model chain will be superimposed to the input chain.
    """
    return _fetch_with_chain(chain, fetch_swiss_model)


from crimm.StructEntities.Atom import Atom, DisorderedAtom
from crimm.StructEntities.TopoElements import Bond, Angle, Dihedral, Improper, CMap
from crimm.StructEntities.Chain import Chain, PolymerChain, Heterogens
from crimm.StructEntities.Residue import Residue, Heterogen, DisorderedResidue
from crimm.StructEntities.TopoDefinitions import ResidueDefinition, AtomDefinition, PatchDefinition
from crimm.StructEntities.Structure import Structure
from crimm.StructEntities.Model import Model






import os
import tempfile
import pycharmm as chm
from pycharmm import read, write, psf, coor
from pycharmm.psf import delete_atoms
from pycharmm.select_atoms import SelectAtoms
from pycharmm import generate
from pycharmm import ic, cons_harm
from pycharmm import minimize as _minimize
from crimm.IO import get_pdb_str

nucleic_letters_1to3 = {
    'A': 'ADE', 'C': 'CYT', 'G': 'GUA', 'T': 'THY', 'U': 'URA',
}

def load_topology(topo_loader, append = False):
    with tempfile.NamedTemporaryFile('w', encoding = "utf-8") as tf:
        tf.write('*RTF Loaded from crimm\n')
        for line in topo_loader._raw_data_strings:
            if line.upper().startswith('RESI') or line.upper().startswith('PRES'):
                line = '\n'+line
            tf.write(line+'\n')
        tf.flush() # has to flush first for long files!
        read.rtf(tf.name, append = append)
        
def load_parameters(param_loader, append = False):
    with tempfile.NamedTemporaryFile('w', encoding = "utf-8") as tf:
        tf.write('*PRM Loaded from crimm\n')
        for line in param_loader._raw_data_strings:
            tf.write(line+'\n')
        tf.flush()
        read.prm(tf.name, append = append, flex=True)

def load_chain(chain, hbuild = False, report = False):
    if not chain.is_continuous():
        raise ValueError("Chain is not continuous! Fix the chain first!")
    segid = f'PRO{chain.id[0]}'
    m_chain = _get_charmm_named_chain(chain, segid)
    first_res = m_chain.child_list[0]
    last_res = m_chain.child_list[-1]
    if (patch_name:=first_res.topo_definition.patch_with) is not None:
        first_patch = patch_name
    else:
        first_patch = ''
        
    if (patch_name:=last_res.topo_definition.patch_with) is not None:
        last_patch = patch_name
    else:
        last_patch = ''
        
    with tempfile.NamedTemporaryFile('w') as tf:
        tf.write(get_pdb_str(m_chain))
        tf.write('END\n')
        tf.flush()
        _load_chain(
            tf.name, segnames=[segid], first_patch=first_patch, last_patch=last_patch,
            hbuild = False
        )
        
def _get_charmm_named_chain(chain, segid):
    m_chain = chain.copy()
    if chain.chain_type == 'Polypeptide(L)':
        for res in m_chain:
            res.segid = segid
            if res.resname == 'HIS':
                res.resname = 'HSD'
    elif chain.chain_type == 'Polyribonucleotide':
        for res in m_chain:
            res.segid = segid
            res.resname = nucleic_letters_1to3[res.resname]
    return m_chain


def _load_chain(
        pdb_filepath, segnames, hbuild = True, report = False,
        first_patch='', last_patch=''
    ):
     
    read.sequence_pdb(pdb_filepath)

    for seg in segnames:
        generate.new_segment(
            seg_name=seg, 
            first_patch=first_patch,
            last_patch=last_patch,
            setup_ic=True,
        )

    read.pdb(pdb_filepath, resi=True)

    ic.prm_fill(replace_all=False)
    ic.build()

    if report:
        chm.lib.charmm.ic_print()
    if hbuild:
        chm.lingo.charmm_script("hbuild sele type H* end")

def minimize(
        constrained_atoms='CA', 
        sd_nstep=1000,
        abnr_nstep=500,
    ):

    cons_harm.setup_absolute(
        selection=chm.SelectAtoms(atom_type=constrained_atoms),
        force_constant=50
        )
    if int(sd_nstep) > 0:
        _minimize.run_sd(nstep=int(sd_nstep))
    else:
        print('Steepest-descend minimization not performed')
    if int(abnr_nstep) > 0:
        _minimize.run_abnr(nstep=int(abnr_nstep), tolenr=1e-3, tolgrd=1e-3)
    else:
        print('Adopted Basis Newton-Raphson minimization not performed')
    cons_harm.turn_off()
    
def ok_to_sync(chain):
    resname_list = []
    if chain.undefined_res:
        print(f"Undefined residue exists! {chain.undefined_res}")
        return False
    for res in chain:
        if res.topo_definition is not None:
            resname_list.append(res.topo_definition.resname)
        else:
            return False
    return resname_list == psf.get_res()

def sync_coords(chain):
    if not ok_to_sync(chain):
        print("[Abort] Possible residue sequences mismatch!")
        return
    ibase = list(psf.get_ibase())
    new_coord_df = coor.get_positions()
    atom_coords = list(zip(psf.get_atype(), new_coord_df.to_numpy()))
    for i, (st, end) in enumerate(zip(ibase[:-1], ibase[1:])):
        cur_res = chain.residues[i]
        for atom_name, coordinate in atom_coords[st:end]:
            if atom_name in cur_res:
                cur_res[atom_name].coord = coordinate
    print(f'Synchronized: {chain}')
    
def remove_all_atoms():
    all_atoms = SelectAtoms(select_all=True)
    delete_atoms(selection=all_atoms)


import warnings
import json
import requests
from rdkit import Chem
from rdkit.Chem import AllChem
from crimm.IO import get_pdb_str

class LigandBondOrderException(Exception):
    """Define class LigandBondOrderException."""

class SmilesQueryException(Exception):
    """Define class SmilesQueryException."""

class LigandBondOrderWarning(Warning):
    """Define class LigandBondOrderException."""

class SmilesQueryWarning(Warning):
    """Define class SmilesQueryException."""

def _build_smiles_query(resname):
    query = '''
    {
        chem_comps(comp_ids: ["{var_lig_id}"]) {
            rcsb_id
            rcsb_chem_comp_descriptor {
            SMILES
            }
        }
    }
    '''.replace("{var_lig_id}", resname)
    return query

def query_rcsb_for_smiles(het_res):
    """Query the canonical SMILES for the heterogen molecule from PDB based on chem_comps
    ID"""
    url="https://data.rcsb.org/graphql"
    response = requests.post(
            url, json={'query': _build_smiles_query(het_res.resname)}, timeout=1000
        )
    response_dict = json.loads(response.text)
    return_vals = response_dict['data']['chem_comps']
    if not return_vals:
        warnings.warn(
            f'Query on {het_res.resname} did not return requested information: '
            'chem_comps (smiles string)'
        )
        return

    smiles = return_vals[0]['rcsb_chem_comp_descriptor']['SMILES']
    return smiles

def heterogen_to_rdkit(het_res, smiles=None):
    """Convert a heterogen to an rdkit mol. The bond orders are set based on the
    SMILES string. If the SMILES string is not provided, the function will query
    the RCSB PDB for the SMILES string.
    
    Args:
        het_res (Heterogen): The heterogen to be converted.
        smiles (str, optional): The SMILES string of the heterogen. Defaults to None.
        
    Returns:
        rdkit.Chem.rdchem.Mol: The rdkit mol of the heterogen.
    """
    mol = Chem.MolFromPDBBlock(get_pdb_str(het_res))
    if smiles is None:
        smiles = query_rcsb_for_smiles(het_res)
    # We do not allow rdkit mol return if the correct bond orders are not set
    if smiles is None:
        msg = (
            'Fail to set bond orders on the Ligand mol! PDB query on SMILES does not'
            'return any result.'
        )
        warnings.warn(msg, SmilesQueryWarning)
        return

    template = AllChem.MolFromSmiles(smiles)
    template = Chem.RemoveHs(template)
    try:
        mol = AllChem.AssignBondOrdersFromTemplate(template, mol)
        if het_res.pdbx_description is not None:
            mol.SetProp('Description', str(het_res.pdbx_description))
        return mol

    except ValueError:
        msg = (
            'No structure match found! Possibly the SMILES string supplied or reported on PDB '
            'mismatches the ligand structure.'
        )
        warnings.warn(msg, LigandBondOrderWarning)
        return



import warnings
import numpy as np
from Bio.Align import PairwiseAligner
from Bio.PDB import Superimposer
from crimm.StructEntities.Chain import PolymerChain
from crimm.Visualization.NGLVisualization import show_nglview_multiple

##TODO: Refactor this!!
class ChainSuperimposer(Superimposer):
    """Superimpose two chains using their canonical sequences if available"""
    def __init__(self) -> None:
        super().__init__()
        self.ref_chain: PolymerChain = None
        self.mov_chain: PolymerChain = None
        self.aligner = PairwiseAligner()
        # We don't want highly fragmented alignments
        self.aligner.target_internal_open_gap_score = -100
        self.aligner.query_internal_open_gap_score = -100
        self.alignments = None
        self.ref_ranges, self.mov_ranges = None, None

    def _check_chain_type(self, chain):
        if not isinstance(chain, PolymerChain):
            raise TypeError(
                "ChainSuperimposer only works with PolymerChain class"
            )
        
    def get_matching_res(self, ref_chain, mov_chain):
        """ Find all aligned residues that exist in both chains """
        self.ref_chain = ref_chain
        self.mov_chain = mov_chain
        if self.ref_chain.can_seq == self.mov_chain.can_seq:
            # If canonical sequences are identical
            # Use all present residues from both chains
            self.alignments = ["identical sequences"]
            complete_range = range(1, len(self.ref_chain.can_seq)+1)
            id_pairs = zip(complete_range, complete_range)

        else:
            # Canonical sequences are not identical
            # Align them first to find the common segments
            self.alignments = self.aligner.align(
                self.ref_chain.can_seq, self.mov_chain.can_seq
            )
            top_alignment = self.alignments[0]
            self.ref_ranges, self.mov_ranges = top_alignment.aligned
            res_ids = top_alignment.indices+1 # residue seq ids start from 1
            aligned_ids = np.logical_and(res_ids[0], res_ids[1])
            id_pairs = res_ids.T[aligned_ids]

        r_res, m_res = self._get_aligned_res(id_pairs)
        return r_res, m_res

    def _get_aligned_res(self, id_pairs):
        """
        Get a set of residue ids from the aligned ranges
        """
        ref_aligned_res = []
        mov_aligned_res = []
        for ref_id, mov_id in id_pairs:
            # needs to be converted to int because the ids are numpy.int64
            ref_id, mov_id = int(ref_id), int(mov_id)
            if ref_id in self.ref_chain and mov_id in self.mov_chain:
                ref_res = self.ref_chain[ref_id]
                mov_res = self.mov_chain[mov_id]
                if ref_res.resname != mov_res.resname:
                    warnings.warn(
                        "Residues are not identical at "
                        f"model chain: {ref_res.resname}-{ref_res.id[1]} and "
                        f"template chain: {mov_res.resname}-{mov_res.id[1]}"
                    )
                    continue
                ref_aligned_res.append(ref_res)
                mov_aligned_res.append(mov_res)

        return ref_aligned_res, mov_aligned_res

    def set_chains(self, ref_chain, mov_chain, on_atoms = 'CA'):
        """Set the chains to be superimposed. mov_chain will be moved to
        ref_chain."""

        self._check_chain_type(ref_chain)
        self._check_chain_type(mov_chain)

        seq_len = len(ref_chain.can_seq)
        # We don't want highly fragmented alignments. Gap opening is only allowed 
        # in the middle of the sequence when 20% or more of the sequence length 
        # is aligned as a result of that
        self.aligner.target_internal_open_gap_score = -(seq_len // 5)
        self.aligner.query_internal_open_gap_score = -(seq_len // 5)

        ref_res, mov_res = self.get_matching_res(ref_chain, mov_chain)

        if on_atoms == 'CA':
            ref_align_atoms, mov_align_atoms = \
                    self._find_common_CA_atoms(ref_res, mov_res)
        elif on_atoms == 'backbone':
            ref_align_atoms, mov_align_atoms = \
                    self._find_common_backbone_atoms(ref_res, mov_res)
        elif on_atoms == 'all':
            ref_align_atoms, mov_align_atoms = \
                    self._find_all_common_atoms(ref_res, mov_res)
        else:
            raise ValueError('on_atoms has to be selected from '
                '{"backbone", "all", "CA"}') 
        self.set_atoms(ref_align_atoms, mov_align_atoms)

    def _find_all_common_atoms(self, ref_res_list, mov_res_list):
        """Find all common atoms from two lists of residues that are aligned"""
        ref_align_atoms = []
        mov_align_atoms = []
        for ref_res, mov_res in zip(ref_res_list, mov_res_list):
            for atom_id in ref_res.child_dict.keys():
                # Make sure move chain residues have the same atoms
                # Otherwise, the missing atom will be excluded
                if atom_id in mov_res:
                    ref_align_atoms.append(ref_res[atom_id])
                    mov_align_atoms.append(mov_res[atom_id])

        return ref_align_atoms, mov_align_atoms

    def _find_common_CA_atoms(self, ref_res_list, mov_res_list):
        ref_align_atoms = []
        mov_align_atoms = []
        for ref_res, mov_res in zip(ref_res_list, mov_res_list):
            # Make sure both residues have the CA atom
            # Otherwise, the missing atom will be excluded
            if 'CA' in ref_res and 'CA' in mov_res:
                ref_align_atoms.append(ref_res['CA'])
                mov_align_atoms.append(mov_res['CA'])

        return ref_align_atoms, mov_align_atoms

    def _find_common_backbone_atoms(self, ref_res_list, mov_res_list):
        ref_align_atoms = []
        mov_align_atoms = []
        for ref_res, mov_res in zip(ref_res_list, mov_res_list):
            for atom_id in ['N','CA','C','O']:
                # Make sure both residues have the backbone atom
                # Otherwise, the missing atom will be excluded
                if atom_id in ref_res and atom_id in mov_res:
                    ref_align_atoms.append(ref_res[atom_id])
                    mov_align_atoms.append(mov_res[atom_id])

        return ref_align_atoms, mov_align_atoms

    def find_valid_edge_ids(self, gap_ids, avail_res, cutoff):
        """Find the residue id range that is present around the gap."""
        avail_ids = [res.id[1] for res in avail_res]
        left, right = min(gap_ids), max(gap_ids)
        res_ids = []
        self._recur_find_ids(avail_ids, cutoff, left, res_ids, -1)
        self._recur_find_ids(avail_ids, cutoff, right, res_ids, 1)
        return sorted(res_ids)

    @staticmethod
    def _recur_find_ids(avail_ids, remainder, cur_id, valid_ids, direction):
        for i in range(1, remainder+1):
            # forward_direction == 1, backward == -1
            cur_id += 1*direction
            if cur_id < 1 or cur_id > max(avail_ids):
                # inclusive on both edges: 1 to max(avail_ids)
                return
            if cur_id in avail_ids:
                valid_ids.append(cur_id)
            else:
                ChainSuperimposer._recur_find_ids(
                    avail_ids, remainder+1-i, cur_id, valid_ids, direction
                )
                return

    def set_around_gap(
            self, ref_chain, mov_chain,
            ref_gap,
            cutoff=10
        ):

        """
        Set atoms for superposition around a gap/missing residues on the ref 
        chain. Start and end residue index needed. Alignment will be performed on
        backbone atoms. If mov_gap is not specified, identical sequence will be 
        assumed on both chains, and the reference chain gap locations will be applied
        on the move chain 
        """
        ## TODO: add options for on_atoms

        self._check_chain_type(ref_chain)
        self._check_chain_type(mov_chain)
        # Get all matched residues from alignment
        all_ref_res, all_mov_res = self.get_matching_res(ref_chain, mov_chain)
        # Find the existing residue ids that are within the cutoff range
        ref_res_ids = self.find_valid_edge_ids(ref_gap, all_ref_res, cutoff)

        ref_edge_res = []
        mov_edge_res = []
        for ref_res, mov_res in zip(all_ref_res, all_mov_res):
            if ref_res.id[1] in ref_res_ids:
                ref_edge_res.append(ref_res)
                mov_edge_res.append(mov_res)

        if ref_edge_res is None:
            warnings.warn('The cutoff value extends out of the available '
                    'residues in the chain. All backbone atoms on both chains '
                    'are selected for superposition')
            ref_edge_res, mov_edge_res = all_ref_res, all_mov_res

        ref_atoms, mov_atoms = self._find_common_backbone_atoms(
            ref_edge_res, mov_edge_res
        )
        self.set_atoms(ref_atoms, mov_atoms)
    
    def apply_transform(self, entity: PolymerChain=None):
        if entity is None:
            entity = self.mov_chain
        self.apply(entity.get_atoms())

    def show(self):
        show_nglview_multiple([self.ref_chain, self.mov_chain])